# risc.rb - Ruby Internet Service Checker (main library)
#
# written by Taku YASUI
# Copyright (C) Taku YASUI <tach@debian.or.jp>

require 'getoptlong'
require 'thread'
require 'ping'
require 'net/telnet'
require 'risc/services'

class Risc

  NAME = 'risc'
  VERSION = '0.3.9'

  class Config
    def initialize(filename='risc.conf')
      @services = []
      @plugins = []
      @options = {}
      @logtype = nil
      @logoptions = nil
      @storetype = nil
      @storeoptions = nil
      @concurrency = 5
      @plugindir = '/usr/share/risc/plugins'
      begin
        instance_eval(File.readlines(filename).join)
      rescue => e
	STDERR.puts("Failed to read config file: #{e}")
	exit 1
      end
    end
    attr_reader :services, :concurrency, :logtype, :logoptions, :storetype, :storeoptions, :plugins, :options, :plugindir
  end

  def initialize(configfile='risc.conf')
    @after_check_hooks = []
    @conf = Risc::Config.new(configfile)
    begin
      require "risc/log/#{@conf.logtype}"
    rescue LoadError
      require 'risc/log'
    end
    @log = Risc::Log.new(@conf.logoptions)
    begin
      require "risc/store/#{@conf.storetype}"
    rescue LoadError
      require 'risc/store'
    end
    @store = Risc::Store.new(@conf.storeoptions)
    @conf.plugins.each do |plugin|
      begin
        instance_eval(File.readlines("#{@conf.plugindir}/#{plugin}.rb").to_s)
      rescue Exception
        @log.write_err("Cannot load plugin: #{plugin}")
      end
    end
  end

  def check_ping(host)
    begin
      return true if ( Ping.pingecho(host) )
      @conf.services[host].each do |service|
        return true if ( Ping.pingecho(host, 5, service[0]) )
      end
      return false
    rescue Exception
      return false
    end
  end

  def check_service(service)
    begin
      t = Net::Telnet.new({ 'Host' => service[0],
                            'Port' => service[1],
      	  		    'Telnetmode' => true })
    rescue Exception
      return false
    end
    begin
      return true if ( ! service[2] && ! service[3] )
      if ( service[2] )
        t.puts(service[2])
      end
      if ( service[3] )
        t.waitfor(service[3])
      end
    rescue Exception
      t.close
      return false
    end
    t.close
    return true
  end

  def add_after_check_hook(block=Proc::new)
    @after_check_hooks << block
  end

  def after_check_hook
    @after_check_hooks.each do |proc|
      proc.call
    end
  end

  def run
    m = Mutex.new
    q = Queue.new
    threads = []

    @conf.concurrency.times do |c|
      threads.push(Thread.new(c) do |t|
        while ( service = q.pop ) do
	  t.exit unless ( service )
	  if ( check_service(service) )
	    m.synchronize do
              @log.write("#{service[0]}:#{service[1]} is alive.")
	      @store.save_alive(service[0], service[1])
	    end
	  else
	    m.synchronize do
              @log.write_err("#{service[0]}:#{service[1]} seems to be down!")
	      @store.save_dead(service[0], service[1])
	    end
	  end
        end
      end)
    end

    @conf.services.each_key do |host|
      if ( ! check_ping(host) )
	m.synchronize do
          @log.write_err("#{host} is down!")
	  @store.save_dead(host)
        end
        next
      end
      m.synchronize do
        @log.write("#{host} is running.")
        @store.save_alive(host)
      end
      @conf.services[host].each do |service|
        q.push([host, *service])
      end
    end
    @conf.concurrency.times { q.push(false) }

    threads.each { |t| t.join }

    after_check_hook
  end

  attr_reader :store, :conf

end

Risc.new.run if ( $0 == __FILE__ )
